@@ -13,7 +13,7 @@
  *      Anatoly Vorobey <mellon@pobox.com>
  *      Brad Fitzpatrick <brad@danga.com>
  *
- *  $Id: memcached.c,v 1.28 2003/07/22 09:14:33 bradfitz Exp $
+ *  $Id: memcached.c,v 1.42 2003/10/08 19:25:37 bradfitz Exp $
  */
 
 #include "config.h"
@@ -23,6 +23,12 @@
 #include <sys/socket.h>
 #include <sys/signal.h>
 #include <sys/resource.h>
+/* some POSIX systems need the following definition
+ * to get mlockall flags out of sys/mman.h.  */
+#ifndef _P1003_1B_VISIBLE
+#define _P1003_1B_VISIBLE
+#endif
+#include <pwd.h>
 #include <sys/mman.h>
 #include <fcntl.h>
 #include <stdlib.h>
@@ -30,13 +36,17 @@
 #include <string.h>
 #include <unistd.h>
 #include <netinet/in.h>
+#include <netinet/tcp.h>
 #include <arpa/inet.h>
 #include <errno.h>
 #include <time.h>
 #include <event.h>
-#include <malloc.h>
 #include <assert.h>
 
+#ifdef HAVE_MALLOC_H
+#include <malloc.h>
+#endif
+
 #include "memcached.h"
 
 struct stats stats;
@@ -87,6 +97,14 @@ conn **freeconns;
 int freetotal;
 int freecurr;
 
+void set_cork (conn *c, int val) {
+    if (c->is_corked == val) return;
+    c->is_corked = val;
+#ifdef TCP_NOPUSH
+    setsockopt(c->sfd, IPPROTO_TCP, TCP_NOPUSH, &val, sizeof(val));
+#endif
+}
+
 void conn_init(void) {
     freetotal = 200;
     freecurr = 0;
@@ -125,6 +143,13 @@ conn *conn_new(int sfd, int init_state, int event_flags) {
         stats.conn_structs++;
     }
 
+    if (settings.verbose > 1) {
+        if (init_state == conn_listening)
+            fprintf(stderr, "<%d server listening\n", sfd);
+        else
+            fprintf(stderr, "<%d new client connection\n", sfd);
+    }
+
     c->sfd = sfd;
     c->state = init_state;
     c->rlbytes = 0;
@@ -140,6 +165,8 @@ conn *conn_new(int sfd, int init_state, int event_flags) {
     c->write_and_free = 0;
     c->item = 0;
 
+    c->is_corked = 0;
+
     event_set(&c->event, sfd, event_flags, event_handler, (void *)c);
     c->ev_flags = event_flags;
 
@@ -158,6 +185,9 @@ void conn_close(conn *c) {
     /* delete the event, the socket and the conn */
     event_del(&c->event);
 
+    if (settings.verbose > 1)
+        fprintf(stderr, "<%d connection closed.\n", c->sfd);
+
     close(c->sfd);
 
     if (c->item) {
@@ -200,6 +230,9 @@ void conn_close(conn *c) {
 void out_string(conn *c, char *str) {
     int len;
 
+    if (settings.verbose > 1)
+        fprintf(stderr, ">%d %s\n", c->sfd, str);
+
     len = strlen(str);
     if (len + 2 > c->wsize) {
         /* ought to be always enough. just fail for simplicity */
@@ -314,6 +347,8 @@ void process_stat(conn *c, char *command) {
         return;
     }
 
+#ifdef HAVE_MALLOC_H
+#ifdef HAVE_STRUCT_MALLINFO
     if (strcmp(command, "stats malloc") == 0) {
         char temp[512];
         struct mallinfo info;
@@ -333,6 +368,8 @@ void process_stat(conn *c, char *command) {
         out_string(c, temp);
         return;
     }
+#endif /* HAVE_STRUCT_MALLINFO */
+#endif /* HAVE_MALLOC_H */
 
     if (strcmp(command, "stats maps") == 0) {
         char *wbuf;
@@ -448,6 +485,13 @@ void process_command(conn *c, char *command) {
      * directly into it, then continue in nread_complete().
      */ 
 
+    if (settings.verbose > 1)
+        fprintf(stderr, "<%d %s\n", c->sfd, command);
+
+    /* All incoming commands will require a response, so we cork at the beginning,
+       and uncork at the very end (usually by means of out_string)  */
+    set_cork(c, 1);
+
     if ((strncmp(command, "add ", 4) == 0 && (comm = NREAD_ADD)) || 
         (strncmp(command, "set ", 4) == 0 && (comm = NREAD_SET)) ||
         (strncmp(command, "replace ", 8) == 0 && (comm = NREAD_REPLACE))) {
@@ -698,7 +742,7 @@ int try_read_network(conn *c) {
         if (c->rbytes >= c->rsize) {
             char *new_rbuf = realloc(c->rbuf, c->rsize*2);
             if (!new_rbuf) {
-                if(settings.verbose)
+                if (settings.verbose > 0)
                     fprintf(stderr, "Couldn't realloc input buffer\n");
                 c->rbytes = 0; /* ignore what we read */
                 out_string(c, "SERVER_ERROR out of memory");
@@ -747,32 +791,33 @@ void drive_machine(conn *c) {
     int res;
 
     while (!exit) {
-      /*printf("state %d\n", c->state); */
+        /* printf("state %d\n", c->state);*/
         switch(c->state) {
         case conn_listening:
             addrlen = sizeof(addr);
             if ((sfd = accept(c->sfd, &addr, &addrlen)) == -1) {
                 if (errno == EAGAIN || errno == EWOULDBLOCK) {
-                    perror("accept() shouldn't block");
+                    exit = 1;
+                    break;
                 } else {
                     perror("accept()");
                 }
-                return;
+                break;
             }
             if ((flags = fcntl(sfd, F_GETFL, 0)) < 0 ||
                 fcntl(sfd, F_SETFL, flags | O_NONBLOCK) < 0) {
                 perror("setting O_NONBLOCK");
                 close(sfd);
-                return;
+                break;
             }            
             newc = conn_new(sfd, conn_read, EV_READ | EV_PERSIST);
             if (!newc) {
-                if(settings.verbose)
+                if (settings.verbose > 0)
                     fprintf(stderr, "couldn't create new connection\n");
                 close(sfd);
-                return;
+                break;
             }
-            exit = 1;
+
             break;
 
         case conn_read:
@@ -784,7 +829,7 @@ void drive_machine(conn *c) {
             }
             /* we have no command line and no data to read from network */
             if (!update_event(c, EV_READ | EV_PERSIST)) {
-                if(settings.verbose)
+                if (settings.verbose > 0)
                     fprintf(stderr, "Couldn't update event\n");
                 c->state = conn_closing;
                 break;
@@ -825,7 +870,7 @@ void drive_machine(conn *c) {
             }
             if (res == -1 && (errno == EAGAIN || errno == EWOULDBLOCK)) {
                 if (!update_event(c, EV_READ | EV_PERSIST)) {
-                    if(settings.verbose) 
+                    if (settings.verbose > 0) 
                         fprintf(stderr, "Couldn't update event\n");
                     c->state = conn_closing;
                     break;
@@ -834,7 +879,7 @@ void drive_machine(conn *c) {
                 break;
             }
             /* otherwise we have a real error, on which we close the connection */
-            if(settings.verbose)
+            if (settings.verbose > 0)
                 fprintf(stderr, "Failed to read, and not due to blocking\n");
             c->state = conn_closing;
             break;
@@ -870,7 +915,7 @@ void drive_machine(conn *c) {
             }
             if (res == -1 && (errno == EAGAIN || errno == EWOULDBLOCK)) {
                 if (!update_event(c, EV_READ | EV_PERSIST)) {
-                    if(settings.verbose)
+                    if (settings.verbose > 0)
                         fprintf(stderr, "Couldn't update event\n");
                     c->state = conn_closing;
                     break;
@@ -879,7 +924,7 @@ void drive_machine(conn *c) {
                 break;
             }
             /* otherwise we have a real error, on which we close the connection */
-            if(settings.verbose)
+            if (settings.verbose > 0)
                 fprintf(stderr, "Failed to read, and not due to blocking\n");
             c->state = conn_closing;
             break;
@@ -892,6 +937,8 @@ void drive_machine(conn *c) {
                     c->write_and_free = 0;
                 }
                 c->state = c->write_and_go;
+                if (c->state == conn_read)
+                    set_cork(c, 0);
                 break;
             }
             res = write(c->sfd, c->wcurr, c->wbytes);
@@ -903,7 +950,7 @@ void drive_machine(conn *c) {
             }
             if (res == -1 && (errno == EAGAIN || errno == EWOULDBLOCK)) {
                 if (!update_event(c, EV_WRITE | EV_PERSIST)) {
-                    if(settings.verbose)
+                    if (settings.verbose > 0)
                         fprintf(stderr, "Couldn't update event\n");
                     c->state = conn_closing;
                     break;
@@ -913,7 +960,7 @@ void drive_machine(conn *c) {
             }
             /* if res==0 or res==-1 and error is not EAGAIN or EWOULDBLOCK,
                we have a real error, on which we close the connection */
-            if(settings.verbose)
+            if (settings.verbose > 0)
                 fprintf(stderr, "Failed to write, and not due to blocking\n");
             c->state = conn_closing;
             break;
@@ -937,7 +984,7 @@ void drive_machine(conn *c) {
                 }
                 if (res == -1 && (errno == EAGAIN || errno == EWOULDBLOCK)) {
                     if (!update_event(c, EV_WRITE | EV_PERSIST)) {
-                        if(settings.verbose)
+                        if (settings.verbose > 0)
                             fprintf(stderr, "Couldn't update event\n");
                         c->state = conn_closing;
                         break;
@@ -947,7 +994,7 @@ void drive_machine(conn *c) {
                 }
                 /* if res==0 or res==-1 and error is not EAGAIN or EWOULDBLOCK,
                    we have a real error, on which we close the connection */
-                if(settings.verbose)
+                if (settings.verbose > 0)
                     fprintf(stderr, "Failed to write, and not due to blocking\n");
                 c->state = conn_closing;
                 break;
@@ -977,6 +1024,8 @@ void drive_machine(conn *c) {
                     it = *(c->icurr);
                     assert((it->it_flags & ITEM_SLABBED) == 0);
                     sprintf(c->ibuf, "VALUE %s %u %u\r\n", ITEM_key(it), it->flags, it->nbytes - 2);
+                    if (settings.verbose > 1)
+                        fprintf(stderr, ">%d sending key %s\n", c->sfd, ITEM_key(it));
                     c->iptr = c->ibuf;
                     c->ibytes = strlen(c->iptr);
                     c->ipart = 1;
@@ -1008,7 +1057,7 @@ void event_handler(int fd, short which, void *arg) {
 
     /* sanity */
     if (fd != c->sfd) {
-        if(settings.verbose)
+        if (settings.verbose > 0)
             fprintf(stderr, "Catastrophic: event fd doesn't match conn fd!\n");
         conn_close(c);
         return;
@@ -1052,6 +1101,9 @@ int server_socket(int port) {
     setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, &flags, sizeof(flags));
     setsockopt(sfd, SOL_SOCKET, SO_KEEPALIVE, &flags, sizeof(flags));
     setsockopt(sfd, SOL_SOCKET, SO_LINGER, &ling, sizeof(ling));
+#if !defined(TCP_NOPUSH)
+    setsockopt(sfd, IPPROTO_TCP, TCP_NODELAY, &flags, sizeof(flags));
+#endif
 
     addr.sin_family = AF_INET;
     addr.sin_port = htons(port);
@@ -1112,10 +1164,12 @@ void usage(void) {
     printf("-p <num>      port number to listen on\n");
     printf("-l <ip_addr>  interface to listen on, default is INDRR_ANY\n");
     printf("-d            run as a daemon\n");
+    printf("-u <username> assume identity of <username> (only when run as root)\n");
     printf("-m <num>      max memory to use for items in megabytes, default is 64 MB\n");
     printf("-c <num>      max simultaneous connections, default is 1024\n");
     printf("-k            lock down all paged memory\n");
     printf("-v            verbose (print errors/warnings while in event loop)\n");
+    printf("-vv           very verbose (also print client commands/reponses)\n");
     printf("-h            print this help and exit\n");
     printf("-i            print memcached and libevent license\n");
     return;
@@ -1200,12 +1254,15 @@ int main (int argc, char **argv) {
     struct in_addr addr;
     int lock_memory = 0;
     int daemonize = 0;
+    char *username = 0;
+    struct passwd *pw;
+    struct sigaction sa;
 
     /* init settings */
     settings_init();
 
     /* process arguments */
-    while ((c = getopt(argc, argv, "p:m:c:khivdl:")) != -1) {
+    while ((c = getopt(argc, argv, "p:m:c:khivdl:u:")) != -1) {
         switch (c) {
         case 'p':
             settings.port = atoi(optarg);
@@ -1226,7 +1283,7 @@ int main (int argc, char **argv) {
             lock_memory = 1;
             break;
         case 'v':
-            settings.verbose = 1;
+            settings.verbose++;
             break;
         case 'l':
             if (!inet_aton(optarg, &addr)) {
@@ -1239,6 +1296,9 @@ int main (int argc, char **argv) {
         case 'd':
             daemonize = 1;
             break;
+        case 'u':
+            username = optarg;
+            break;
         default:
             fprintf(stderr, "Illegal argument \"%c\"\n", c);
             return 1;
@@ -1253,19 +1313,50 @@ int main (int argc, char **argv) {
     conn_init();
     slabs_init(settings.maxbytes);
 
+    /* lock paged memory if needed */
+    if (lock_memory) {
+#ifdef HAVE_MLOCKALL
+        mlockall(MCL_CURRENT | MCL_FUTURE);
+#else
+        fprintf(stderr, "warning: mlockall() not supported on this platform.  proceeding without.\n");
+#endif
+    }
+
+    /* lose root privileges if we have them */
+    if (getuid()== 0 || geteuid()==0) {
+        if (username==0 || *username=='\0') {
+            fprintf(stderr, "can't run as root without the -u switch\n");
+            return 1;
+        }
+        if ((pw = getpwnam(username)) == 0) {
+            fprintf(stderr, "can't find the user %s to switch to\n", username);
+            return 1;
+        }
+        if (setgid(pw->pw_gid)<0 || setuid(pw->pw_uid)<0) {
+            fprintf(stderr, "failed to assume identity of user %s\n", username);
+            return 1;
+        }
+    }
 
     if (daemonize) {
         int res;
-        res = daemon(0, 0);
+        res = daemon(0, settings.verbose);
         if (res == -1) {
             fprintf(stderr, "failed to daemon() in order to daemonize\n");
             return 1;
         }
     }
 
-    /* lock paged memory if needed */
-    if (lock_memory) {
-        mlockall(MCL_CURRENT | MCL_FUTURE);
+    /*
+     * ignore SIGPIPE signals; we can use errno==EPIPE if we
+     * need that information
+     */
+    sa.sa_handler = SIG_IGN;
+    sa.sa_flags = 0;
+    if (sigemptyset(&sa.sa_mask) == -1 ||
+        sigaction(SIGPIPE, &sa, 0) == -1) {
+        perror("failed to ignore SIGPIPE; sigaction");
+        exit(1); 
     }
 
     /* create the listening socket and bind it */
