@@ -1,8 +1,12 @@
 /* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
-/* $Id: memcached.h,v 1.16 2003/07/22 09:14:33 bradfitz Exp $ */
+/* $Id: memcached.h,v 1.19 2003/08/13 00:05:19 bradfitz Exp $ */
 
 #define DATA_BUFFER_SIZE 2048
 
+#if defined(TCP_CORK) && !defined(TCP_NOPUSH)
+#define TCP_NOPUSH TCP_CORK
+#endif
+
 struct stats {
     unsigned int  curr_items;
     unsigned int  total_items;
@@ -89,6 +93,7 @@ typedef struct {
     int    wbytes; 
     int    write_and_go; /* which state to go into after finishing current write */
     void   *write_and_free; /* free this memory after finishing writing */
+    char    is_corked;         /* boolean, connection is corked */
 
     char   *rcurr;
     int    rlbytes;
@@ -113,7 +118,7 @@ typedef struct {
     item   **icurr;
     int    ileft;
     int    ipart;     /* 1 if we're writing a VALUE line, 2 if we're writing data */
-    char   ibuf[256]; /* for VALUE lines */
+    char   ibuf[300]; /* for VALUE lines */
     char   *iptr;
     int    ibytes;
                          
